""" Definition of the python side NTuple class.

These are mainly io routines. The idea is: if the file format is the
native numpy format, read of write directly, else pipe binary data
through ntcat.


"""

import numpy as np
from numpy.lib.utils import safe_eval
import re
import sys
import copy
from glob import glob
import os

def stringify(nt):
    '''Cast a record array so that byte fields become utf8 fields'''
    dt = [(f, t.replace('S', 'U')) for f, t in  nt.dtype.descr]
    return nt.astype(dt)

def read_header(fh):
    keys = {}
    names = []
    for i, line in enumerate(fh):
        if line.startswith('#'):
            if line[1:].strip() == "end":
                break
            else:
                names.append(line[1:].split(':')[0].strip())
        elif line.startswith('@'):
            l = line[1:].split()
            keys[l[0]] = convert(' '.join(l[1:]))
    return keys, names
    
def read_header_keys(filename):
    with open(filename) as fh:
        keys, names = read_header(fh)
    return keys, names

def select(ntuple, condition, vars=[], vars_rule="and"):
    """Provide an easier way to express condition on fields

    ga = nacl.util.io.select(nt_atmosphere, "(LAMBDA>5000.) & (LAMBDA<7000.) ")
    
    Works for more than one keyword

    vars is a list of  indexes over which to expand the condition string if 
    built like:
    condition = "field_%s < 1.", vars=["1", "3"]
    This will apply the condition on field_1 and field_3. Logical and.

    vars_rule is set to loop over vars with a & operator
    Set it to "or" "OR" or "|" to get an | operator. 
    Defaults to & for everything else

    Note that vars only loops is %s is provided.
    condition = "field_%d < 1" % (my_index)  works while providing my_index to vars won't work.


    In case of ntuple keys with a dot "target.name" for example,
    the field in the condition should be written with two underscores instead.

    "target__name == 'LSQ10vw'" will be understood as looking for the key target.name 
    (Actually, select() reads the keys with dots replacing them by __). 
    What this means is that if your key is target_name, things will also work
    fine.

    It is if and only if you have both target.name and target__name as keys, that
    you are screwed. But then, maybe that's also when you should start wondering
    if you are not asking for it a little bit, don't you agree ?

    """
    fields = dict([(k.replace(".", "__"), ntuple[k]) for k in ntuple.dtype.names])
    if not vars:
        index = eval(condition, globals(), fields)
    else:
        n = condition.count('%s')
        index = eval(condition % ((vars[0],) * n), globals(), fields)
        for var in vars[1:]:
            if vars_rule in ["or", "OR", "|"]:
                index |= eval(condition % ((var,) * n), globals(), fields)
            else:
                index &= eval(condition % ((var,) * n), globals(), fields)                
                
    return ntuple[index]

def sanitize(name):
    for s in '-*+/ ':
        name = name.replace(s, '_')
    return name

def eval_form(ntuple, formula, var=""):
    """ Provide an easier way to evaluate formula on fields

    For example, if the ntuple has fields named MAG_U and ZP_U
    we can do
    >>> nacl.util.io.eval_form(ntuple, "MAG_%s - ZP_%s", var="U")        
    """
    fields = dict([(sanitize(k), ntuple[k]) for k in ntuple.dtype.names])
    n = formula.count('%s')
    if n:
        formula = formula % ((var,) * n)
    return eval(formula, globals(), fields)


def read_matrix(lines):
    s = np.loadtxt(lines[0:1])
    m = np.loadtxt(lines[1:])
    assert m.shape == (s[0], s[1]), ('Unable to read matrix at'
                                     ' the end of the file')
    return m

# I want to be able to add a field to a NTuple without having to
# remember where those guys hid the method
from numpy.lib.recfunctions import append_fields, join_by, merge_arrays
from numpy.lib.recfunctions import rec_append_fields, rec_drop_fields


def rec_stack(*args, **keys):
    """ Easy concatenation of several record arrays into a single one

    Optionally rename the fields by adding a suffix or a prefix (or
    both) to the field names.

    Exemple:
    --------
    >>> nt0 = np.rec.fromarrays([range(3)], names=['a'])
    >>> nt1 = np.rec.fromarrays([range(3)], names=['a'])
    >>> rec_stack(nt0, nt1, suffix1="_1")
    array([(0, 0), (1, 1), (2, 2)],
          dtype=[('a', '<i8'), ('a_1', '<i8')])

    Recipes:
    --------
    + If you have a number of similar ntuples, for example in a dictionnary 
    >>> d{"label_1":{"nt_description":nt_1, "more info":stuff}, {"nt_description":nt_1, "more info":stuff}  }
    >>> l_nt = [ d[key]["nt_description"] for key in d.keys() ]
    >>> d_suffix = {"suffix%d"%(i) : "_"+label for i, label in enumerate(d.keys())}
    >>> nacl.util.io.rec_stack(*l_nt, **d_suffix)

    will create the NTuple where all the ntuple fields that would have had the same name are now of the form
    same-name_label, where label is the key that was used in the dictionnary.

    """
    merge_result = []
    for i, nt in enumerate(args):
        suffix = keys.pop('suffix%d' % i, "")
        prefix = keys.pop('prefix%d' % i, "")
        nt = nt.copy()
        # This ensure that a new dtype object is created and that the
        # change of field name is not going to affect the original
        # array
        nt.dtype = np.dtype(nt.dtype.descr)
        if suffix or prefix:
            nt.dtype.names = tuple([prefix + name + suffix
                                    for name in nt.dtype.names])
        merge_result.append(nt)
    return merge_arrays(merge_result, flatten=True)


def convert(value):
    value = value.strip()
    if not value:
        value = "nan"
    try:
        value = int(value)
    except ValueError:
        try:
            value = float(value)
        except ValueError:
            pass
    return value


def convert_format(value, format):
    value = value.strip()
    if format[0] == "I":
        if not value:
            value = -99
        else:
            value = int(value)
    elif format[0] in "FE":
        if not value:
            value = np.nan
        else:
            value = float(value)
    else:
        if not value:
            value = '_'
    return value



def find_closest(data, target, sorted=True):
    """
    Stolen from stakoverflow to find index closest to target from a sorted list A.

    If sorted is False, takes care of it
    """
    
    if not sorted:
        idx_sorted = np.argsort(data)
        A = data[idx_sorted]
    else:
        A = data
        idx_sorted = np.arange(len(data) - 1)
        
    idx = A.searchsorted(target)
    idx = np.clip(idx, 1, len(A)-1)
    left = A[idx-1]
    right = A[idx]
    idx -= target - left < right - target
    if idx > len(idx_sorted) -1:
        idx = -1
    return idx_sorted[idx]

class split_container:
    """
    This is a dummy class to create namespaces to contain the data provided by
    nacl.util.io.split method  
    
    >>>  c = croaks.croaks_split_namespace()
    >>> nacl.util.io.split(ntuple, c.__dict__)    
    """
    pass


def split(ntuple, d_output, two_D={}, field_output_suffix="_field"):
    """
    Parameters:
    -----------
    + ntuple:  a nacl.util.io.NTuple
    + d_output: is a dictionnary that will contain the splitted ntuple
    Note that d_output can be 
        - a genuine dictionnary
        - a class instance internal dictionnary c.__dict__
            This allows to access the splitted ntuple data via c.LAMBDA for
            example, if the ntuple has such a vector column
        - the global namespace dictionnary global()
            This allows to access the splitted ntuple data via LAMBDA for
            example if the ntuple has such a vector column
            WARNING: the ntuple should not have names that overwrite anything
                     in the global namespace. Ensuring this is of the user's
                     responsibility 
    + two_D: a dictionnary containing the 2D array names and the re regexp to
             do the grouping
             For example:
             two_D = {"Flux": "FLUX_.*"}
             will try to group all the columns with names matching the regexp
             in a 2D array that will be stored in d_output["Flux"].
             
             If the number of matching column names is N_col, and the NTuple
             length is N_row, the result will be an (N_row x N_col) array.

    + field_output_suffix: 

    Definition:
    -----------
    This method works on "horizontal" NTuples, i.e. of the form:
    LAMBDA  FLUX_label_1 FLUX_label_2 GA_label_1  GA_label_2 GA_3  V1  V2

    LAMBDA, V1 and V3 are different from FLUX_* and GA_* in that they are
    vectors that give a global information on FLUX_* and GA_*, while all FLUX_*
    for example are expected to be compatible (i.e., operations like calculating
    the mean make sense on them).
    
    nacl.util.io.split will by default return each column of the NTuple as a
    vector in d_output, of the form d_output[column_name] = colum_vector_value.

    In the case of our example:
    d["LAMBDA"] = wavelength_vector
    

    If two_D dictionnary is used, the fields corresponding to the regexp
    provided will be agregated in 2D arrays with the same orientation as what
    could be read in the NTuple ascii file.

    Examples:
    ---------
    For an NTuple ntuple  of the form:
    LAMBDA  FLUX_label_1 FLUX_label_2 GA_label_1  GA_label_2 GA_3  V1  V2
    
    with N_l lines

    >>> nacl.util.io.split(ntuple, globals(), two_D={"flux": "FLUX_.*"})
    will add in the global name space:
    
    >>> LAMBDA
    Which will be the vector ntuple["LAMBDA"]
    
    >>> flux
    Which will be the 2D array with
    
    >>> flux[:,0]
    corresponding to ntuple["FLUX_label_1"]

    >>> flux_field[0]
    corresponding to "FLUX_label_1"

    >>> flux[:, 0] - ntuple[flux_field[0]]
    should therefore return only zeroes. 
    

    It can also be useful to declare
    >>> class c:
    >>>    pass
    
    And then do
    >>> nacl.util.io.split(ntuple, globals(), two_D={"flux": "FLUX_.*"})

    This has the advantage of protecting the global namespace from polution. The
    data nacl.util.io.split from the NTuple being the accessible with things like

    >>> c.LAMBDA
    >>> c.flux[:,0]

    It is then easy to calculate sums, means and rms on the 2D arrays with
    things like
    >>> c.flux.mean(axis=1)
    to get the average flux spectrum
   
    """
    # list of ntuple fields, that will be consumed
    # as we deal with the fields
    l_ntuple_field = list(ntuple.dtype.names)

    # Looks for columns to group in 2D arrays:
    for key_two_D in list(two_D.keys()):
        l_field = []
        l_data = []        
        for ntuple_field in l_ntuple_field:
            if re.search(two_D[key_two_D], ntuple_field):
                # the ntuple_field is popped out in order to only leave vectors
                # once this loop is over
                l_data.append(ntuple[ntuple_field])
                l_field.append(ntuple_field)
            else:
                continue
        if len(l_field) != 0:
            # The .T is to inforce the Universal Matrix Convention
            d_output[key_two_D] = np.vstack(l_data).T
            d_output[key_two_D + field_output_suffix] = l_field

            # Update l_ntuple_field by popping out the used values:
            for field in l_field:
                l_ntuple_field.remove(field)
            
    for ntuple_field in l_ntuple_field:
        d_output[ntuple_field] = ntuple[ntuple_field]


class FortranDecription():
    """ Class to handle Fortran ReadMe files of the kind:
    
    --------------------------------------------------------------------------------
   Bytes Format  Units   Label    Explanations
--------------------------------------------------------------------------------
   1-  4  I4     ---     HR       [1/9110]+ Harvard Revised Number
                                    = Bright Star Number
   5- 14  A10    ---     Name     Name, generally Bayer and/or Flamsteed name
  15- 25  A11    ---     DM       Durchmusterung Identification (zone in
                                    bytes 17-19)
  26- 31  I6     ---     HD       [1/225300]? Henry Draper Catalog Number
  32- 37  I6     ---     SAO      [1/258997]? SAO Catalog Number
  38- 41  I4     ---     FK5      ? FK5 star Number
      42  A1     ---     IRflag   [I] I if infrared source
      43  A1     ---   r_IRflag  *[ ':] Coded reference for infrared source
      44  A1     ---    Multiple *[AWDIRS] Double or multiple-star code
  45- 49  A5     ---     ADS      Aitken's Double Star Catalog (ADS) designation
  50- 51  A2     ---     ADScomp  ADS number components
  52- 60  A9     ---     VarID    Variable star identification
  61- 62  I2     h       RAh1900  ?Hours RA, equinox B1900, epoch 1900.0 (1)
  63- 64  I2     min     RAm1900  ?Minutes RA, equinox B1900, epoch 1900.0 (1)
  65- 68  F4.1   s       RAs1900  ?Seconds RA, equinox B1900, epoch 1900.0 (1)
      69  A1     ---     DE-1900  ?Sign Dec, equinox B1900, epoch 1900.0 (1)
  70- 71  I2     deg     DEd1900  ?Degrees Dec, equinox B1900, epoch 1900.0 (1)
  72- 73  I2     arcmin  DEm1900  ?Minutes Dec, equinox B1900, epoch 1900.0 (1)
  74- 75  I2     arcsec  DEs1900  ?Seconds Dec, equinox B1900, epoch 1900.0 (1)
  76- 77  I2     h       RAh      ?Hours RA, equinox J2000, epoch 2000.0 (1)
  78- 79  I2     min     RAm      ?Minutes RA, equinox J2000, epoch 2000.0 (1)
  80- 83  F4.1   s       RAs      ?Seconds RA, equinox J2000, epoch 2000.0 (1)
      84  A1     ---     DE-      ?Sign Dec, equinox J2000, epoch 2000.0 (1)
  85- 86  I2     deg     DEd      ?Degrees Dec, equinox J2000, epoch 2000.0 (1)
  87- 88  I2     arcmin  DEm      ?Minutes Dec, equinox J2000, epoch 2000.0 (1)
  89- 90  I2     arcsec  DEs      ?Seconds Dec, equinox J2000, epoch 2000.0 (1)
  91- 96  F6.2   deg     GLON     ?Galactic longitude (1)
  97-102  F6.2   deg     GLAT     ?Galactic latitude (1)
 103-107  F5.2   mag     Vmag     ?Visual magnitude (1)
     108  A1     ---   n_Vmag    *[ HR] Visual magnitude code
     109  A1     ---   u_Vmag     [ :?] Uncertainty flag on V
 110-114  F5.2   mag     B-V      ? B-V color in the UBV system
     115  A1     ---   u_B-V      [ :?] Uncertainty flag on B-V
 116-120  F5.2   mag     U-B      ? U-B color in the UBV system
     121  A1     ---   u_U-B      [ :?] Uncertainty flag on U-B
 122-126  F5.2   mag     R-I      ? R-I   in system specified by n_R-I
     127  A1     ---   n_R-I      [CE:?D] Code for R-I system (Cousin, Eggen)
 128-147  A20    ---     SpType   Spectral type
     148  A1     ---   n_SpType   [evt] Spectral type code
 149-154  F6.3 arcsec/yr pmRA    *?Annual proper motion in RA J2000, FK5 system
 155-160  F6.3 arcsec/yr pmDE     ?Annual proper motion in Dec J2000, FK5 system
     161  A1     ---   n_Parallax [D] D indicates a dynamical parallax,
                                    otherwise a trigonometric parallax
 162-166  F5.3   arcsec  Parallax ? Trigonometric parallax (unless n_Parallax)
 167-170  I4     km/s    RadVel   ? Heliocentric Radial Velocity
 171-174  A4     ---   n_RadVel  *[V?SB123O ] Radial velocity comments
 175-176  A2     ---   l_RotVel   [<=> ] Rotational velocity limit characters
 177-179  I3     km/s    RotVel   ? Rotational velocity, v sin i
     180  A1     ---   u_RotVel   [ :v] uncertainty and variability flag on
                                    RotVel
 181-184  F4.1   mag     Dmag     ? Magnitude difference of double,
                                    or brightest multiple
 185-190  F6.1   arcsec  Sep      ? Separation of components in Dmag
                                    if occultation binary.
 191-194  A4     ---     MultID   Identifications of components in Dmag
 195-196  I2     ---     MultCnt  ? Number of components assigned to a multiple
     197  A1     ---     NoteFlag [*] a star indicates that there is a note
                                    (see file notes)
"""
    def __init__(self):
        self.fields = []
        self.bytes = []
        self.format = []
        self.unit = []
        self.Explanation = []
        
    def read_descr(self, description):
        for line in description.splitlines():
            if not line:
                continue
            entries = line.split()
            entries = [e.strip() for e in entries]
            if entries[0][0] not in "1234567890":
                continue
            entries.reverse()
            byterange = entries.pop().split('-')
            if len(byterange) == 1:
                self.bytes.append((int(byterange[0]) - 1, int(byterange[0])))
            elif len(byterange) == 2 and not byterange[1]:
                self.bytes.append((int(byterange[0]) - 1, int(entries.pop())))
            elif len(byterange) == 2 and byterange[1]:
                self.bytes.append((int(byterange[0]) - 1, int(byterange[1])))
            else:
                raise ValueError('Line not understood : ' + line)
            self.format.append(entries.pop())
            self.unit.append(entries.pop())
            self.fields.append(entries.pop())
            entries.reverse()
            self.Explanation.append(' '.join(entries))

    def read_file(self, fname):
        with open(fname) as fid:
            entries = []
            for line in fid:
                entries.append([convert_format(line[a:b], form) for (a, b), form in zip(self.bytes, self.format)])
        return entries

    def read_fid(self, fid):
        entries = []
        for line in fid:
            entries.append([convert_format(line[a:b], form) for (a, b), form in zip(self.bytes, self.format)])
        return entries

def find_fortran_description(filename, tablename):
    with open(filename) as fid:
        for line in fid:
            if line.startswith('Byte-by-byte Description of file: %s' % tablename):
                break
        descr = []
        for line in fid:
            if line.startswith('Byte-by-byte Description of file'):
                break
            descr.append(line)
    return ''.join(descr)

def readdescription(fid):
    descr = []
    sep = 0
    while (sep < 4):
        descr.append(fid.readline())
        if descr[-1].startswith("--"):
            sep += 1
    return "".join(descr)


class NTuple(np.ndarray):
    """ Add Toads-like file formatting to np record arrays

    Exemples
    --------
    >>> nt = NTuple.fromfile('filename.ntuple')
    >>> nt['col2'] += 1
    >>> nt.keys['modified'] = "by me"
    >>> nt.tofile('newfilename.ntuple')

    One may convert a classic np record array to benefit from the
    io function using the view mecanism:
    >>> a = array(zip(randn(10), randn(10)), {'names':['x','y'],'formats':('float',)*2})
    >>> b = a.view(NTuple)
    >>> b.tofile('filename.ntuple')
    """
    def __new__(subtype, shape, dtype=float, buffer=None, offset=0,
                strides=None, order=None, keys={}):
        obj = np.ndarray.__new__(subtype, shape, dtype, buffer,
                                    offset, strides, order)
        obj.keys = keys
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.keys = getattr(obj, 'keys', {})

    def __call__(self, *args, **keys):
        return select(self, *args, **keys)
        
    @classmethod
    def fromfile(cls, f, mmap_mode=None):
        own_fid = False
        if isinstance(f, str):
            fid = open(f, "rb")
            own_fid = True
        else:
            fid = f

        try:
            # try to distinguish from Np binary files and txt files.
            N = len(np.lib.format.MAGIC_PREFIX)
            magic = fid.read(N)
            fid.seek(-N, 1)  # back-up
            if magic == np.lib.format.MAGIC_PREFIX:
                if mmap_mode:
                    return open_memmap(f, mode=mmap_mode)
                else:
                    a, keys = read_array(fid)
                    a = a.view(NTuple)
                    a.keys = keys
                    return a
            else:
                if mmap_mode:
                    raise IOError("mmap mode not supported for txt files")
                if not own_fid:
                    raise IOError("f must be a filename, cannot read ascii"
                                  " data from existing file handle")
                import subprocess
                p = subprocess.Popen(['ntcat', '-b', f],
                                     stdout=subprocess.PIPE, bufsize=4096)
                a, keys = read_array(p.stdout)
                a = a.view(NTuple)
                a.keys = keys
                p.wait()
                return a
        finally:
            if own_fid:
                fid.close()

    def tofile(self, f, format=None):
        own_fid = False
        if isinstance(f, str):
            fid = open(f, "wb")
            own_fid = True
            if format is None:
                if f.endswith('.npy') or f.endswith('.nxt'):
                    format = "binary"
                else:
                    format = "txt"
        else:
            fid = f
            if format is None:
                format = "binary"
        try:
            if format == "binary":
                write_array(fid, self)
            else:
                import subprocess
                p = subprocess.Popen(['ntcat', '-'], stdout=fid,
                                     stdin=subprocess.PIPE, bufsize=4096)
                write_array(p.stdin, self)
                p.stdin.close()
                p.wait()
        finally:
            if own_fid:
                fid.close()
                
    def keys_to_list(self, key_string, dtype=None):
        key_string = self.keys[key_string]
        key_string = key_string.replace("[", "")
        key_string = key_string.replace("]", "")
        
        key_string = key_string.strip()
        l_key = key_string.split(",")
        
        if dtype is None:
            return np.array(l_key)
        else:
            return np.array(l_key, dtype=dtype)

                
    def totxt(self, filename=None):
        """
        Write the tuple to a *.list fileformat, with values that can
        be something else than doubles

        If filename is not provided, will write to stdout

        @key1 value
        @key2 value
        #value_name_1: description
        #value_name_2: description
        #end
        value1 value2
        value1 value2        
        """

        if filename is None:
            fh = sys.stdout
        else:
            fh = open(filename, "w")

        for key, value in self.keys.items():
            print("@%s %s" % (key, str(value)), file=fh)

        for name in self.dtype.names:
            print("#%s:" % (name), file=fh)
        print("#end", file=fh)

        tmp_line = ""
        for t_line in self:
            for name in self.dtype.names:
                tmp_line += "%-s  " % (str(t_line[name]))
            tmp_line += "\n"
        print(tmp_line, file=fh)
        if hasattr(self, 'm'):
            print("#MATRIX", file=fh)
            print("%d %d" % self.m.shape, file=fh)
            np.savetxt(fh, self.m)

    @classmethod
    def fromtxt(self, filename):
        comments = set(['#', '\n'])
        with open(filename, "r") as fid:
            keys, names = read_header(fid)
            records = []
            for line in fid:
                if line[0] in comments:
                    continue
                vals = line.split()
                records.append([convert(v) for v in vals])
        nt = np.rec.fromrecords(records, names=names).view(NTuple)
        nt.keys = keys
        return nt
    
    @classmethod
    def fromtxtslow(self, filename):
        """
        Read the tuple from a *.list fileformat, with values that can
        be something else than doubles

        @key1 value
        @key2 value
        #value_name_1: description
        #value_name_2: description
        #end
        value1 value2
        value1 value2
        """
        fh = open(filename, "r")

        d_key = {}
        l_name = []

        hasm = False
        l_line = []
        lines = fh.readlines()
        for i, line in enumerate(lines):
            if line.startswith('#MATRIX'):
                m = read_matrix(lines[i + 1:])
                hasm = True
                break
            if re.search("^#", line):
                if re.search(":", line):
                    line = line.split(":")[0]
                    line = re.sub("#", "", line)
                    line = re.sub("\s*", "", line)
                    line = line.strip()
                    if line == "end":
                        continue
                    else:
                        l_name.append(line)
                else:
                    #- a # without : means a comment
                    continue
            elif re.search("^@", line):
                group = re.search("@(.+?)\s+(.+)", line)
                group = group.groups()
                try:
                    d_key[group[0]] = float(group[1])
                except:
                    d_key[group[0]] = group[1]
            else:
                l_line.append(line)
        fh.close()

        l_value = []
        #- Check that the file contains actual data, not only a header
        if len(l_line) != 0:
            for line in l_line:
                #- remove comments
                line = line.split("#")[0]
                #- skip empty lines
                if re.search("^\s+$", line):
                    continue
                else:
                    line = line.strip()
                    line = re.sub("\s+", " ", line)
                    tmp_l_value = re.split("\s", line)
                    for (i_v, value) in enumerate(tmp_l_value):
                        try:
                            tmp_l_value[i_v] = int(value)
                        except ValueError:
                            try:
                                tmp_l_value[i_v] = float(value)
                            except ValueError:
                                continue
                    l_value.append(tmp_l_value)
            l_value = np.rec.fromrecords(l_value, names=l_name)
            l_value = l_value.view(NTuple)
            l_value.keys = d_key
            if hasm:
                l_value.m = m
        else:  # If the file has no actual data, we make a dummy empty recarray
            l_value = np.recarray((0,),
                                     dtype=[(name, "S1") for name in l_name])
            l_value = l_value.view(NTuple)
            l_value.keys = d_key
        return l_value

    @classmethod
    def fromsnana(self, filename):
        """Read the tuple from a *.dat snana format file.

        KEY1: value
        KEY2: value
        VARLIST: name1 name2 ...
        OBS: value1 value2
        OBS: value1 value2
        END:
        """
        fh = open(filename, "r")

        d_key = {}
        names = []
        hasm = False
        records = []
        
        for i, line in enumerate(fh):
            entries = line.split()
            if entries[0] == 'VARLIST:':
                names = entries[1:]
            elif entries[0] == "OBS:":
                records.append([convert(v) for v in entries[1:]])
            elif entries[0] == "END:":
                break
            else:
                d_key[entries[0][:-1]] = " ".join(entries[1:])
        fh.close()
        nt = np.rec.fromrecords(records, names=names)
        nt = nt.view(NTuple)
        nt.keys = d_key
        return nt

    @classmethod
    def fromorg(self, filename):
        fh = open(filename, "r")

        d_key = {}
        names = []
        records = []
        first = True        
        header_expr = re.compile(':\s*(.+)\s*:\s*(.+)\s*')
        
        lines = fh.readlines()
        fh.close()
        for i, line in enumerate(lines):
            if line.startswith(':'):
                try:
                    key, val = header_expr.match(line).groups()
                except:
                    raise ValueError('Line not understood : ' + line)
                try:
                    val = float(val)
                except:
                    pass
                d_key[key] = val
            if line.startswith('|-'):
                continue
            if line.startswith('|'):
                vals = line.split('|')[1:-1]
                if first:
                    names = [v.strip() for v in vals]
                    first = False
                else:
                    records.append([convert(v) for v in vals])

        nt = np.rec.fromrecords(records, names=names)
        nt = nt.view(NTuple)
        nt.keys = d_key
        return nt

    def toorg(self, filename=None):
        """
        Write the tuple to a .org file. 

        Parameters
        ----------
          - filename [str]: output file name 
                            if None, write to stdout instead

        """
        if filename is None:
            fh = sys.stdout
        else:
            fh = open(filename, 'w')
            
        # global keys 
        for key, value in self.keys.items():
            print(":%s: %s" % (key, str(value)), file=fh)
        
        # header 
        ncols = len(self.keys)
        line_sep = "|" + "-+"*(ncols-1) + "-|"
        print(line_sep, file=fh) 
        s = "| "
        for key in self.dtype.names:
            s += key ; s += " | "
        print(s, file=fh)
        print(line_sep, file=fh)
        
        # data 
        tmp_line = ""
        for t_line in self:
            for name in self.dtype.names:
                tmp_line += '| %-s ' % (str(t_line[name]))
            tmp_line += '|\n'
        print(tmp_line[:-1], file=fh)
        print(line_sep, file=fh) 
    
    @classmethod
    def fromcsv(self, filename):
        fh = open(filename, "r")

        d_key = {}
        names = []
        records = []
        first = True

        lines = fh.readlines()
        fh.close()
        for i, line in enumerate(lines):
            if line.startswith('%'):
                continue
            vals = line.split(',')
            if first:
                names = [v.strip() for v in vals]
                first = False
            else:
                records.append([convert(v) for v in vals])

        nt = np.rec.fromrecords(records, names=names)
        nt = nt.view(NTuple)
        nt.keys = d_key
        return nt

    @classmethod
    def fromfortran(self, filename, description):
        f = FortranDecription()
        f.read_descr(description)
        if isinstance(filename, str):
            nt = f.read_file(filename)
        else:
            nt = f.read_fid(filename)
        nt = np.rec.fromrecords(nt, names=f.fields)
        nt = nt.view(NTuple)
        nt.keys = {}
        return nt

# from numpy.lib.format
def _wrap_header(header, version):
    """
    Takes a stringified header, and attaches the prefix and padding to it
    """
    _header_size_info = {(1, 0): ('<H', 'latin1'), (2, 0): ('<I', 'latin1'), (3, 0): ('<I', 'utf8')}
    import struct
    assert version is not None
    fmt, encoding = _header_size_info[version]
    if not isinstance(header, bytes):  # always true on python 3
        header = header.encode(encoding)
    hlen = len(header) + 1
    padlen = np.lib.format.ARRAY_ALIGN - ((np.lib.format.MAGIC_LEN + struct.calcsize(fmt) + hlen) % np.lib.format.ARRAY_ALIGN)
    try:
        header_prefix = np.lib.format.magic(*version) + struct.pack(fmt, hlen + padlen)
    except struct.error:
        msg = "Header length {} too big for version={}".format(hlen, version)
        raise ValueError(msg)

    # Pad the header with spaces and a final newline such that the magic
    # string, the header-length short and the header are aligned on a
    # ARRAY_ALIGN byte boundary.  This supports memory mapping of dtypes
    # aligned up to ARRAY_ALIGN on systems like Linux where mmap()
    # offset must be page-aligned (i.e. the beginning of the file).
    return header_prefix + header + b' '*padlen + b'\n'

    
def write_array_header_1_0(fp, d):
    """ Write the header for an array using the 1.0 format.

    Parameters
    ----------
    fp : filelike object
    d : dict
        This has the appropriate entries for writing its string representation
        to the header of the file.
    """
    import struct
    header = ["{"]
    for key, value in sorted(d.items()):
        # Need to use repr here, since we eval these when reading
        header.append("'%s': %s, " % (key, repr(value)))
    header.append("}")
    header = "".join(header)
    header = _wrap_header(header, version=(1,0))
    fp.write(header)

def write_array(fp, array, version=(1, 0)):
    """
    Write an array to an NPY file, including a header.

    If the array is neither C-contiguous nor Fortran-contiguous AND the
    file_like object is not a real file object, this function will have to
    copy data in memory.

    Parameters
    ----------
    fp : file_like object
        An open, writable file object, or similar object with a ``.write()``
        method.
    array : ndarray
        The array to write to disk.
    version : (int, int), optional
        The version number of the format.  Default: (1, 0)

    Raises
    ------
    ValueError
        If the array cannot be persisted.
    Various other errors
        If the array contains Python objects as part of its dtype, the
        process of pickling them may raise various errors if the objects
        are not picklable.

    """
    if version != (1, 0):
        msg = "we only support format version (1,0), not %s"
        raise ValueError(msg % (version,))
    # magic number now handled in write_array_header_1_0
    #    fp.write(np.lib.format.magic(*version))
    write_array_header_1_0(fp, header_data_from_ntuple_1_0(array))
    array = np.asarray(array)
    if array.dtype.hasobject:
        # We contain Python objects so we cannot write out the data directly.
        # Instead, we will pickle it out with version 2 of the pickle protocol.
        cPickle.dump(array, fp, protocol=2)
    elif array.flags.f_contiguous and not array.flags.c_contiguous:
        if isinstance(fp, str):#file):
            fp.write(array.T.tostring('C'))
        else:
            array.T.tofile(fp)

    else:
        if isinstance(fp, str):
            fp.write(array.tostring('C'))
        else:
            # XXX: We could probably chunk this using something like
            # arrayterator.
            array.tofile(fp)


def header_data_from_ntuple_1_0(array):
    """ Get the dictionary of header metadata from a np.ndarray.

    Parameters
    ----------
    array : np.ndarray

    Returns
    -------
    d : dict
        This has the appropriate entries for writing its string representation
        to the header of the file.
    """
    d = {}
    d['shape'] = array.shape
    if array.flags.c_contiguous:
        d['fortran_order'] = False
    elif array.flags.f_contiguous:
        d['fortran_order'] = True
    else:
        # Totally non-contiguous data. We will have to make it C-contiguous
        # before writing. Note that we need to test for C_CONTIGUOUS first
        # because a 1-D array is both C_CONTIGUOUS and F_CONTIGUOUS.
        d['fortran_order'] = False

    d['descr'] = np.lib.format.dtype_to_descr(array.dtype)
    d['keys'] = array.keys
    for k in d['keys']:
        d['keys'][k] = str(d['keys'][k])

    return d


def read_array(fp):
    """
    Read an array from an NPY file.

    Parameters
    ----------
    fp : file_like object
        If this is not a real file object, then this may take extra memory
        and time.

    Returns
    -------
    array : ndarray
        The array from the data on disk.

    Raises
    ------
    ValueError
        If the data is invalid.

    """
    version = np.lib.format.read_magic(fp)
    if version != (1, 0):
        msg = "only support version (1,0) of file format, not %r"
        raise ValueError(msg % (version,))
    shape, fortran_order, dtype, keys = read_array_header_1_0(fp)
    if len(shape) == 0:
        count = 1
    else:
        count = np.multiply.reduce(shape)

    # Now read the actual data.
    if dtype.hasobject:
        # The array contained Python objects. We need to unpickle the data.
        array = cPickle.load(fp)
    else:
        try:
            # We can use the fast fromfile() function.
            array = np.fromfile(fp, dtype=dtype, count=count)
        except TypeError:
            # This is not a real file. We have to read it the memory-intensive
            # way.
            # XXX: we can probably chunk this to avoid the memory hit.
            data = fp.read(int(count * dtype.itemsize))
            array = np.fromstring(data, dtype=dtype, count=count)

        if fortran_order:
            array.shape = shape[::-1]
            array = array.transpose()
        else:
            array.shape = shape

    return array, keys


def read_array_header_1_0(fp):
    """
    Read an array header from a filelike object using the 1.0 file format
    version.

    This will leave the file object located just after the header.

    Parameters
    ----------
    fp : filelike object
        A file object or something with a `.read()` method like a file.

    Returns
    -------
    shape : tuple of int
        The shape of the array.
    fortran_order : bool
        The array data will be written out directly if it is either
        C-contiguous or Fortran-contiguous. Otherwise, it will be made
        contiguous before writing it out.
    dtype : dtype
        The dtype of the file's data.

    Raises
    ------
    ValueError :
        If the data is invalid.

    """
    # Read an unsigned, little-endian short int which has the length of the
    # header.
    import struct
    hlength_str = fp.read(2)
    if len(hlength_str) != 2:
        msg = "EOF at %s before reading array header length"
        raise ValueError(msg % fp.tell())
    header_length = struct.unpack('<H', hlength_str)[0]
    header = fp.read(header_length)
    if len(header) != header_length:
        raise ValueError("EOF at %s before reading array header" % fp.tell())

    # The header is a pretty-printed string representation of a literal Python
    # dictionary with trailing newlines padded to a 16-byte boundary. The keys
    # are strings.
    #   "shape" : tuple of int
    #   "fortran_order" : bool
    #   "descr" : dtype.descr
    #   "keys" : datacards
    try:
        d = safe_eval(header.decode())
    except SyntaxError as e:
        msg = "Cannot parse header: %r\nException: %r"
        raise ValueError(msg % (header, e))
    if not isinstance(d, dict):
        msg = "Header is not a dictionary: %r"
        raise ValueError(msg % d)
    keys = list(d.keys())
    keys.sort()
    if keys != ['descr', 'fortran_order', 'keys', 'shape']:
        if keys != ['descr', 'fortran_order', 'shape']:
            msg = "Header does not contain the correct keys: %r"
            raise ValueError(msg % (keys,))
        else:
            d['keys'] = {}

    # Sanity-check the values.
    if (not isinstance(d['shape'], tuple) or
        not np.all([isinstance(x, int) for x in d['shape']])):
        msg = "shape is not valid: %r"
        raise ValueError(msg % (d['shape'],))
    if not isinstance(d['fortran_order'], bool):
        msg = "fortran_order is not a valid bool: %r"
        raise ValueError(msg % (d['fortran_order'],))
    try:
        dtype = np.dtype(d['descr'])
    except TypeError as e:
        msg = "descr is not a valid dtype descriptor: %r"
        raise ValueError(msg % (d['descr'],))

    return d['shape'], d['fortran_order'], dtype, d['keys']


def open_memmap(filename, mode='r+', dtype=None, shape=None,
                fortran_order=False, version=(1, 0)):
    """
    Open a .npy file as a memory-mapped array.

    This may be used to read an existing file or create a new one.

    Parameters
    ----------
    filename : str
        The name of the file on disk.  This may *not* be a file-like
        object.
    mode : str, optional
        The mode in which to open the file; the default is 'r+'.  In
        addition to the standard file modes, 'c' is also accepted to
        mean "copy on write."  See `memmap` for the available mode strings.
    dtype : data-type, optional
        The data type of the array if we are creating a new file in "write"
        mode, if not, `dtype` is ignored.  The default value is None,
        which results in a data-type of `float64`.
    shape : tuple of int
        The shape of the array if we are creating a new file in "write"
        mode, in which case this parameter is required.  Otherwise, this
        parameter is ignored and is thus optional.
    fortran_order : bool, optional
        Whether the array should be Fortran-contiguous (True) or
        C-contiguous (False, the default) if we are creating a new file
        in "write" mode.
    version : tuple of int (major, minor)
        If the mode is a "write" mode, then this is the version of the file
        format used to create the file.  Default: (1,0)
    Returns
    -------
    marray : memmap
        The memory-mapped array.

    Raises
    ------
    ValueError
        If the data or the mode is invalid.
    IOError
        If the file is not found or cannot be opened correctly.

    See Also
    --------
    memmap

    """
    if not isinstance(filename, str):
        raise ValueError("Filename must be a string.  Memmap cannot use"
                         " existing file handles.")

    if 'w' in mode:
        # We are creating the file, not reading it.
        # Check if we ought to create the file.
        if version != (1, 0):
            msg = "only support version (1,0) of file format, not %r"
            raise ValueError(msg % (version,))
        # Ensure that the given dtype is an authentic dtype object rather than
        # just something that can be interpreted as a dtype object.
        dtype = np.dtype(dtype)
        if dtype.hasobject:
            msg = "Array can't be memory-mapped: Python objects in dtype."
            raise ValueError(msg)
        d = dict(
            descr=dtype_to_descr(dtype),
            fortran_order=fortran_order,
            shape=shape,
        )
        # If we got here, then it should be safe to create the file.
        fp = open(filename, mode + 'b')
        try:
            fp.write(magic(*version))
            write_array_header_1_0(fp, d)
            offset = fp.tell()
        finally:
            fp.close()
    else:
        # Read the header of the file first.
        fp = open(filename, 'rb')
        try:
            version = np.lib.format.read_magic(fp)
            if version != (1, 0):
                msg = "only support version (1,0) of file format, not %r"
                raise ValueError(msg % (version,))
            shape, fortran_order, dtype, keys = read_array_header_1_0(fp)
            if dtype.hasobject:
                msg = "Array can't be memory-mapped: Python objects in dtype."
                raise ValueError(msg)
            offset = fp.tell()
        finally:
            fp.close()

    if fortran_order:
        order = 'F'
    else:
        order = 'C'

    # We need to change a write-only mode to a read-write mode since we've
    # already written data to the file.
    if mode == 'w+':
        mode = 'r+'

    marray = np.memmap(filename, dtype=dtype, shape=shape, order=order,
                          mode=mode, offset=offset)

    return marray, keys


def find_odometer(template):
    existing_files = glob(template +'*.npy')
    existing_files = [int(f.replace('.npy', '').replace(template, '')) for f in existing_files]
    if not existing_files:
        print('Warning: first file in the serie')
        odometer = 0
    else:
        odometer = np.max(existing_files) + 1
    return odometer

class IFile(object):
    def __init__(self, names, filename=None, template='datafile_', freq=10, append=True):
        self.names = names
        self.res = []
        if filename is None:
            odometer = find_odometer(template)
            self.filename = '%s%04d.npy' % (template, odometer)
        else:
            self.filename = filename
            if os.path.exists(filename):
                if append:
                    print('File already exists, appending')
                    tab = np.load(filename)
                    self.res = [[t[n] for n in self.names] for t in tab]
                else:
                    print('Warning: File already exists, it will be overwriten')
        print('Data will be registered in ' + self.filename)
        self.freq = freq
        self.i = 0

    def __enter__(self):
        return self
    
    def __exit__(self, *exec_details):
        self.res = self._write()
                
    def _write(self):
        if self.res:
            tab = np.rec.fromrecords(self.res, names=self.names)
            np.save(self.filename, tab)
            self.i = 0
            return tab
    
    def append(self, *args):
        if len(args) != len(self.names):
            raise ValueError("Incorrect number of arguments. Append should be provided with %s" % str(self.names))
        self.i += 1
        self.res.append(args)
        if self.i == self.freq:
            self._write()
        



"""Spectral libraries and Broadband fluxes

Utility classes to manage collections of spectra. The main purpose of
this code is to compute effiently broadband fluxes from potentially
large series of SEDs.

Rationale
---------
When building broadband fluxes from spectra, it is efficient to first
project the spectra on a unique, well defined spline basis (generally
of low order, e.g.  1 or 2): 

.. math:: S(\lambda) = \sum_i s_i B_i(\lambda) = B^T S
    
If the transmissions functions of the telescope are also developed on
a spline basis (not necessarily the same):

.. math:: T(\lambda) = \sum_i f_i K_i(\lambda) = K^T F

then, the broadband fluxes can be written as:

.. math:: \Phi= \int T(\lambda) \\frac{\lambda}{hc} S(\lambda) d\lambda = F^T G S
        
where :math:`G` is the *gramian* of the two bases, i.e.  the matrix
that contains the integrals of the dyadic products of the basis
elements: :math:`G_{ij} = \int B_i(x) K_j(x) dx`. Since the support of
the splines is finite, :math:`G` is a very sparse band-matrix, and the
computation of the integral above is very fast. In fact, the product
:math:`F^T G` can be precomputed and re-used for each spectrum, hence,
the integral reduces to a scalar product.

Stellar libs and Filter sets
----------------------------

A :class:`StellarLib`, such as :class:`Pickles` or :class:`Calspec`
contains a series of spectra, projected on a B-Spline basis (see e.g.
:class:`bspline.BSpline`).  The class holds an index, which
contains the metadata which happen to be available for each star (e.g.
spectral type, coordinates, ``V``-mag, ``B-V`` color etc.), as well as
a ndarray which contains the basis coefficients for all spectra.

The spectra may be loaded from the original files stored in
``SALTPATH`` and then projected on the basis.  For libraries of
moderate size such as Calspec or Pickles, this takes 1-2
seconds. Alternatively, the index and coefficient arrays may be saved
into a numpy ``.npz`` file (:meth:`StellarLib.save`) and later
reloaded from it (:meth:`StellarLib.load`) much faster, of course.

Examples
--------

Load the calspec library from the original files that are stored in
``SALTPATH``. Project the spectra on a order-2 (i.e. deg-1) BSpline
basis, defined on the range [3000, 12000] divided into 5\AA bins:

>>> import numpy as np
>>> from bspline import BSpline
>>> from . import stellarlibs
>>> grid = np.arange(3000., 12005., 5.)
>>> basis = BSpline(grid, order=2)
>>> calspec = stellarlibs.Calspec(basis=basis)

Save / load the basis into / from a .npz file

>>> calspec.save('calspec.npz')
>>> calspec = stellarlibs.Calspec(filename='calspec.npz')

Prepare a set of filters 

>>> from .instruments import InstrumentModel
>>> grid = np.arange(3000., 12005., 10.)
>>> lsst = InstrumentModel("LSST")
>>> fs = stellarlibs.FilterSet(grid=grid, [lsst.EffectiveFilterByBand(b) for b in "griz"])

Compute broadband fluxes

>>> flx = stellarlibs.flux(calspec, fs)

Compute broadband mags

>>> m = stellarlibs.mag(calspec, fs, "AB")

Load the original spectra (no projections on a basis)

>>> index, spectra = stellarlibs.Pickles.fetch(basis=None)

Integrate the original spectra using Spectrum.IntegFlux

>>> flx = stellarlib.flux(spectra, fs)

Modify the atmosphere in the filter set 

>>> from atmosphericmodel import Buton
>>> atm = Buton()
>>> fs.atmosphere = atm  # multiply by Buton and reproject the filters
>>> flx = stellarlib.flux(spectra, fs)

.. Warning:: IntegFlux and the gram() method (i.e.  the slow and
    fast options) should differ due to the infamous (hc)_snfit vs
    (hs)_scipy problem.  I am correcting the gram() flux by hand to
    adjust them to what IntegFlux returns. At some point, we need to 
    correct both methods. 
"""

import os
import os.path as op
#from exceptions import NotImplementedError, ValueError

import numpy as np
try:
    import pyfits as pf
except ImportError:
    import astropy.io.fits as pf
from .bspline import CardinalBSpline, BSpline, gram, lgram
from .spectrum import Spectrum, flambda_to_fclambda, CLIGHT_A_s, HPLANCK
from .instruments import MagSys
from . import constants
from ..util import saltpath, io

try:
    from sksparse.cholmod import cholesky_AAt
except ImportError:
    from scikits.sparse.cholmod import cholesky_AAt
    
# get rid of this at some point ! 
ALIGN_WITH_INTEG_FLUX = 0.9964833253040145 / 1.001899



class StellarLib(object):
    """Holds a collection of spectra projected on a spline basis.
    
    
    """
    def __init__(self, **kw):
        """Build the library, from a 
        """
        if 'filename' in kw:
            self.load(kw['filename'])
        elif 'basis' in kw:
            self.basis = kw.pop('basis')
            self.grid = self.basis.grid
            self.index, self.fluxes = self.__class__.fetch(self.basis, **kw)
        elif 'grid' in kw:
            self.grid = self.pop('grid')
            self.basis = BSpline(self.grid)
            self.index, self.fluxes = self.__class__.fetch(self.basis, **kw)
        else:
            self.index, self.fluxes = self.__class__.fetch(None,**kw)
            
    def __len__(self):
        return len(self.index)

    def save(self, filename):
        """Save the stellar library contents into a .npz file
        
        Dump the basis, projected spectra and metadata into a single .npz file. 

        Args:
            filename (str): name of the .npz file
        """
        np.savez(filename, grid=self.basis.grid, index=self.index, fluxes=self.fluxes)
        
    def load(self, filename):
        """Load the stellar library from a .npz filename. 

        Load the basis, projected spectra, and spectrum metadata from
        the .npz file specified in argument.

        Args:
            filename (str): name of the .npz file
        """
        r = np.load(filename)
        self.grid = r['grid']
        self.basis = BSpline(self.grid)
        self.index = r['index']
        self.fluxes = r['fluxes']        
        

class Pickles(StellarLib):
    """Spectra from (Pickles, 1992), projected on a BSpline basis.

    This is a specialization of :class:`StellarLib`, for the library
    of spectra described in (Pickles, 1992). This class implements
    essentially a method :meth:`load` which fetches the original
    spectra from their files, and projects them onto the basis
    specified by the user. 
    """
    def __init__(self, **kw):
        """Constructor

        Instantiate the Pickles library.  This is not the original
        library in the sense that all the spectra are projected on a
        BSpline basis (of order 2, unless specified otherwise).

        Args:
            filename (str, optional): load the library from the file 
            basis (BSpline, optional): specify the BSpline basis on
                which to project the spectra
            grid (ndarray, optional): specify the basis grid. An order 2
                (degree 1) BSpline basis will be instantiated on that grid.

        .. TODO 
            add an option to self.load, so that it can retrieve the original 
            (non projected) library.
        """
        super(Pickles, self).__init__(**kw)
        
    @staticmethod
    def fetch(basis):
        """fetch original spectra and project them on the spline basis
        
        Fetch the original spectra (from $SALTPATH/stellarlibs) and
        project them on the basis given in argument.

        Args:
            basis (BSpline type object): basis on which to project the spectra
            
        Returns:
            recarray, ndarray: index, spline coefficients if basis is not None
                               index, original spectra otherwise            
        """
        import glob 
        path = op.join(saltpath.stellar_libs_path, 'pickles')
        # retain only the 'u*.dat' spectra (extended)
        filenames = glob.glob(path + os.sep + 'u*.dat')
        #        filenames = filter(lambda x: op.basename(x)[0]!='u', filenames)
        n = len(filenames)
        
        # data structure 
        # we separate the metadata (put in the index tab)
        # from the spline coefficients (stored in spectra)
        # it seems better to have the spline coeffs contiguous in memory.
        dtype = np.dtype([('styp', '|S10'), 
                         ('filename', '|S20')]) 
        #                ('p', np.float64, len(basis))])
        index  = np.zeros(n, dtype=dtype)
        if basis is not None:
            spectra = np.zeros((n, len(basis)), dtype=np.float64)
            # we oversample a bit the spectra, to overconstrain
            # the basis projection a little
            gx = basis.grid
            ovsp_gx = np.hstack((gx, 0.5*(gx[1:]+gx[:-1])))
            ovsp_gx.sort()
        else:
            spectra = []
        
        # loop on the spectra
        for i,fn in enumerate(filenames):
            with open(fn) as f:
                wl, flux = [], []
                for line in f.readlines():
                    wl.append(float(line[0:7]))
                    flux.append(float(line[8:17]))
            wl, flux = np.asarray(wl), np.asarray(flux)
            index[i]['styp'] = op.basename(fn).split('.')[0]
            index[i]['filename'] = op.basename(fn)            
            
            # oversample the spectral data and project it on the basis
            if basis is not None:
                yy = np.interp(ovsp_gx, wl, flux)
                spectra[i,:] = basis.linear_fit(ovsp_gx, yy)
            else:
                spectra.append(Spectrum(wl, flux))
                
        return index, spectra

class PicklesDepagne(StellarLib):
    """Spectra from (Pickles Depagne, 2010), projected on a BSpline basis.

    This is a specialization of :class:`StellarLib`, for the library
    of spectra described in (Pickles, 1992) plus a few peculiar
    types. The spectral types are matched to the tycho
    photometry. This class implements essentially a method
    :meth:`load` which fetches the original spectra from their files,
    and projects them onto the basis specified by the user.

    """
    def __init__(self, **kw):
        """Constructor

        Instantiate the library.  This is not the original library in
        the sense that all the spectra are projected on a BSpline
        basis (of order 2, unless specified otherwise), and have been
        renormalised to a synthetic mag VT=0.

        Args:
            filename (str, optional): load the library from the file 
            basis (BSpline, optional): specify the BSpline basis on
                which to project the spectra
            grid (ndarray, optional): specify the basis grid. An order 2
                (degree 1) BSpline basis will be instantiated on that grid.
        """
        super(PicklesDepagne, self).__init__(**kw)
        
    @staticmethod
    def fetch(basis):
        """fetch original spectra and project them on the spline basis
        
        Fetch the original spectra (from $SALTPATH/stellarlibs) and
        project them on the basis given in argument.

        Args:
            basis (BSpline type object): basis on which to project the spectra
            
        Returns:
            recarray, ndarray: index, spline coefficients if basis is not None
                               index, original spectra otherwise            
        """
        import glob 
        path = op.join(saltpath.stellar_libs_path, 'picklesdepagne10')
        # retain only the 'u*.dat' spectra (extended)
        filenames = [op.join(path,  '%d.lib' % i) for i in range(141)]
        #        filenames = filter(lambda x: op.basename(x)[0]!='u', filenames)
        n = len(filenames)
        
        # data structure 
        # we separate the metadata (put in the index tab)
        # from the spline coefficients (stored in spectra)
        # it seems better to have the spline coeffs contiguous in memory.
        dtype = np.dtype([('styp', '|S10'), 
                         ('filename', '|S20')]) 
        #                ('p', np.float64, len(basis))])
        index  = np.zeros(n, dtype=dtype)
        if basis is not None:
            spectra = np.zeros((n, len(basis)), dtype=np.float64)
            # we oversample a bit the spectra, to overconstrain
            # the basis projection a little
            gx = basis.grid
            ovsp_gx = np.hstack((gx, 0.5*(gx[1:]+gx[:-1])))
            ovsp_gx.sort()
        else:
            spectra = []
        
        # loop on the spectra
        for i,fn in enumerate(filenames):
            nt = io.NTuple.fromtxt(fn)
            
            index[i]['styp'] = '%d' % i
            index[i]['filename'] = op.basename(fn)            
            
            # oversample the spectral data and project it on the basis
            if basis is not None:
                yy = np.interp(ovsp_gx, nt['wl'], nt['flux'])
                spectra[i,:] = basis.linear_fit(ovsp_gx, yy)
            else:
                spectra.append(Spectrum(nt['wl'], nt['flux']))
                
        return index, spectra

class Calspec(StellarLib):
    """Spectra from the latest CALSPEC release (Bohlin, 2014),
    projected on a BSpline basis.

    This is a specialization of :class:`StellarLib`, for the library
    of spectra described in (Bohlin, 2014). 
    """    
    def __init__(self, **kw):
        super(Calspec, self).__init__(**kw)
            
    @staticmethod
    def fetch(basis, version='2015-01'):
        """load the spectra from the original files
        
        Args:
            basis (BSpline or None): BSpline basis to use
            version (str): CALSPEC release data

        Returns:
            recarray, ndarray: index and spline coeffs if basis is not None
                               index and list of Spectrum.Spectrum
        """
        path = op.join(saltpath.stellar_libs_path,
                       'calspec-' + version)                       
        # read the index 
        index_file = op.join(path, 'index.org')
        if not op.isfile(index_file):
            raise ValueError('unable to find CALSPEC index: %s' % index_file)
        index = io.NTuple.fromorg(index_file)
        
        n = len(index)
        if basis is not None:
            gx = basis.grid
            ovsp_gx = np.hstack((gx, 0.5*(gx[1:]+gx[:-1])))
            ovsp_gx.sort()
            spectra = np.zeros((n, len(basis)), dtype=np.float64)
        else:
            spectra = []
            
        for i,d in enumerate(index):
            with pf.open(path + os.sep + d['FILENAME']) as f:
                wl = f[1].data['WAVELENGTH']
                flux = f[1].data['FLUX']
                if basis is not None:
                    #flux = flambda_to_fclambda(wl, flux)
                    yy = np.interp(ovsp_gx, wl, flux)
                    spectra[i,:] = basis.linear_fit(ovsp_gx, yy)
                else:
                    spectra.append(Spectrum(wl, flux))

        return index, spectra

    
class FilterSet(object):
    """
    Utility class that projects a set of filters on B-Splines
    and stores the projection coefficients. 

    ..  note:: this code should be merged with
        :class:`instruments.FilterWheel`. We keep it here
        until we get to talk with the Great Guardians.    
    """
    def __init__(self, basis, transmissions, atmosphere_model=None, distort_func=None):
        self.basis = basis
        if type(transmissions) is dict:
            self.transmissions = transmissions
        elif type(transmissions) is list:
            self.transmissions = dict([(t.InstrumentName + '::' + t.Band, t) 
                                       for t in transmissions])
        # projection grid (larger than basis grid)
        self._grid, self._J, self._factor = self._projector(self.basis)
        self.atm = np.ones_like(self._grid)
        if distort_func is None:
            self.distort = None
        else:
            self.distort = distort_func(self._grid)
        self.parameters = self._project(basis, self.transmissions)
        self.bands = list(self.transmissions.keys())
        self.mean_wl = dict([(k,t.mean(t.x_min, t.x_max)) for k,t in list(self.transmissions.items())])


    def _projector(self, basis):
        # refine the basis grid
        gx = basis.grid
        gxx = np.hstack((gx, 0.5*(gx[1:]+gx[:-1])))
        gxx.sort()
        # precompute the projector
        # (will save repeated calls to cholesky_AAt)
        J = basis.eval(gxx).tocsr()
        factor = cholesky_AAt(J.T)
        return gxx, J, factor

    def _project(self, basis, transmissions):
        ret = {}
        if type(self.atm) is dict:
            # get rid of this case at some point. 
            # this adds complications to the code
            # we do not need that anymore since 
            # we now can extract a subset of the FilterSet
            for k,t in list(transmissions.items()):
                y = t(self._grid) * self.atm[k]
                if self.distort is not None:
                    y *= self.distort
                ret[k] = self._factor(self._J.T * y)
        else:
            for k,t in list(transmissions.items()):
                y = t(self._grid) * self.atm
                if self.distort is not None:
                    y *= self.distort
                ret[k] = self._factor(self._J.T * y)
        return ret
    
    def __len__(self):
        return len(self.transmissions)

    def __getitem__(self, key):
        return self.parameters[key]

    def extract(self, band_names):
        tr = [self.transmissions[b] for b in band_names]
        return FilterSet(self.basis, tr)

    def mean_wavelength(self, bandnames):
        d = self.mean_wl
        return np.asarray([d[bn] for bn in bandnames])
    
    @property
    def atmosphere(self):
        """return the current instance of the atmospheric transmission.
        
        The atmospheric transmission is stored as a grid vector, of
        size equal to that of the internal grid (self._grid). Return
        that vector.
        """
        return self.atm

    @atmosphere.setter
    def atmosphere(self, atmosphere_model):
        """change the current instance of the atmospheric transmission.
        
        Update the atmospheric transmission vector, and reproject the
        filters.
        
        Args:
            atmosphere_model (callable or dict of callables): 
                 the function that defines the atmospheric transmission.
                 (possibly one different transmission per band)

        .. todo: add atmospheric transmission as an explicit parameter in self._project
        """
        if type(atmosphere_model) is dict:
            # get rid of this at some point 
            self.atm = {}
            for k,f in list(atmosphere_model.items()):
                self.atm[k] = f(self._grid)
        else:
            self.atm = atmosphere_model(self._grid)
        self.parameters = self._project(self.basis, self.transmissions)

    @property 
    def distortion(self):
        return self.distort
    
    @distortion.setter
    def distortion(self, distort_func):
        if distort_func is None:
            self.distort = None
        else:
            self.distort = distort_func(self._grid)
        self.parameters = self._project(self.basis, self.transmissions)


def flux(filters, splib, wl_step=5., z=0.):
    """Compute broadband fluxes (in e-/s) 

    For each filter contained in the filter set, compute the the
    broadband fluxes of all the stars in the spectral library.
    
    The function follows the snfit convention, in the sense that it
    expects *rest frame* SED's, expressed in
    :math:`\mathrm{erg/cm^2/s/A}` and returns *observer frame*
    broadband fluxes expressed in :math:`e^-/s`.  The broadband flux
    is defined as:
            
    .. math:: f = \\frac{C}{(1+z)^2}\int S\\left(\\frac{\lambda}{1+z}\\right)\ \lambda\ T(\lambda)\ d\lambda

    where the integral is computed in the observer frame (hence the
    :math:`1/(1+z)^2` factor).  :math:`T` is the telescope
    transmission, in :math:`e^- cm^2 / \gamma`. The constant C is
    equal to :math:`1/hc` with h in erg s and c in A/s:

    .. math:: \\frac{1}{hc} \\approx 50341170.081942275\ \mathrm{erg^{-1} A^{-1}}

    using the values of h and c in `scipy.constants` (see however the
    warning section below).
        
    Args: 
        splib (inherits from `StellarLib`): a spectral library, or at
            least an object that holds a basis as well as a set of
            coefficients.
        filters (`FilterSet`): a filter set
        wl_step (float, optional): integration step (if using
            traditional method)
        z (float): redshift to apply to the spectrum

    Returns:
        (numpy.rec.array) contains the broadband fluxes. 

    .. note:: The function may be passed either a FilterWheel and a
         list of ``Spectra``, in which case it is going to
         call ``spectrum.IntegFlux`` on each spectrum.  Alternatively,
         we can pass it a ``stellarlibs.FilterSet`` and a
         ``stellarlibs.StellarLib``. In this case, it uses the faster
         Gram method (see above). To decide what to do, it the
         function checks whether splib and filters hold a bspline
         basis or not.

    .. WARNING:: IntegFlux and the gram() method (i.e.  the slow and
         fast options) differ due to the infamous (hc)_snfit vs
         (hs)_scipy problem.  I am correcting the gram() flux by hand
         to adjust them to what IntegFlux returns. At some point, we
         need to correct both methods.
    """
    names, fluxes = [], []
    
    if hasattr(splib, 'basis') and hasattr(filters, 'basis'):
        G = lgram(splib.basis, filters.basis, z=z)
        for key, pars in list(filters.parameters.items()):
            U = G.dot(pars)
            names.append(key)
            fluxes.append(ALIGN_WITH_INTEG_FLUX * np.dot(splib.fluxes, U))
    elif type(splib) is list:
        for k,tr in list(filters.transmissions.items()):
            names.append(k)
            fluxes.append([s.IntegFlux(tr, wavelength_integration_step=wl_step, z=z)
                           for s in splib])
    else:
        raise ValueError("don't know what to do with splib=%s" % splib.__class__)
    
    return np.rec.fromarrays(fluxes, names=names)


# from instruments import MagSys
# class MagSysFS(MagSys):
    
#     def __init__(self, magsys_name, basis=None):
#         super(MagSysFS, self).__init__(magsys_name)
#         if basis is None:
#             grid = np.arange(2000., 12000., 1.)
#             self.basis = BSpline(grid, order=4)
#         print self.pure_ab
#         if self.pure_ab:
#             x = np.arange(2000., 12000., 0.5)
#             self.spectrum = self.basis.linear_fit(x, constants.flux_ab(x))
#         else:
#             self.spectrum = self.
                    
#     def ZeroPoint(self, fs):
#         pass

#     def mag(self, filterset, splib, z=0.):
#         pass
    

def mag(filters, splib, magsys, wl_step=5., z=0.):
    """Broadband mags of the stars in the spectral library. 

    For each filter contained in the filter set, compute the the
    broadband magnitudes, in the specified magnitude system, of all
    the stars in the spectral library.
    
    Args:
        splib (inherits from `StellarLib`): spectral library 
        filters (`FilterSet`): filters to use to compute the mags
        magsys (MagSys): magnitude system to use
        wl_step(float, optional): integration step (if using 
            traditional method).
        z (float): redshift to apply to the spectrum

    Returns:
        (numpy.rec.array): contains the broadband magnitudes.
    """
    names, mags = [], []
    # fast option: splib and filters have a respline() method 
    # that returns the basis and spline coefficients. 
    # we can use the gram() method 
    if hasattr(splib, 'basis') and hasattr(filters, 'basis'):
        G = lgram(splib.basis, filters.basis, z=z)
        for key, pars in list(filters.parameters.items()):
            T = filters.transmissions[key]
            zp = magsys.ZeroPoint(T)
            U = G.dot(pars)
            names.append(key)
            mags.append(-2.5 * np.log10(ALIGN_WITH_INTEG_FLUX * np.dot(splib.fluxes, U)) + zp)
    elif type(splib) is list:
        for key, tr in list(filters.transmissions.items()):
            names.append(key)
            T = filters.transmissions[key]
            zp = magsys.ZeroPoint(T)
            flx = np.asarray([s.IntegFlux(tr, wavelength_integration_step=wl_step)
                              for s in splib])
            mags.append(-2.5 * np.log10(flx) + zp)
    else:
        raise ValueError("don't know what to do with splib=%s" % splib.__class__)
            
            
    return np.rec.fromarrays(mags, names=names)




"""
psf.py 

psf models
"""

import os

import numpy as np 

from numpy.polynomial.polynomial import polyvander2d, polyval2d
from numpy.polynomial.legendre import leggauss
from scipy.interpolate import interp1d, RectBivariateSpline

from ..util.io import NTuple
import scipy.interpolate
from ..lib.fitparameters import FitParameters
from ..lib.instruments import InstrumentModel, MagSys
# TODO: move load_ascii into util or io
from ..lib.spectrum import Spectrum, load_ascii

import logging
logging.basicConfig(format="%(asctime)-15s %(levelname)s %(message)s",
                    level=logging.INFO)
    

def integ_in_pixels(x, y, f, quadrature_order=4):
    """Integrate a 2D function in pixels
    
    Given a function f (typically, a psf profile), and a series of
    positions (x,y), evaluate the integral of the function on the
    pixels corresponding to each of the positions (x,y) specified in
    argument.

    Args
       x (ndarray of floats): x-positions, in pixels.
       y (ndarray of floats): y-positions, in pixels.
       f (callable): function to integrate
       quadrature_order (int, optional): order of the gaussian quadrature to use.

    Returns:
       ndarray of floats: integrated PSF value.

    .. note: 
       the function is written to preserve the shape of x and y
    """
    # put this integration code in cache
    p,w = leggauss(quadrature_order)
    p *= 0.5 ; w *= 0.5
    gx,gy = np.meshgrid(p,p)
    wx,wy = np.meshgrid(w,w)
    ww = (wx*wy).ravel()
    #    xc = (np.floor(x)+0.5).reshape(x.shape + (1,)) + p
    #    yc = (np.floor(y)+0.5).reshape(y.shape + (1,)) + p
    xc = x.reshape(x.shape + (1,1,)) + gx
    yc = y.reshape(y.shape + (1,1,)) + gy
    v = f(xc,yc)
    v = v.reshape(v.shape[0], v.shape[1], -1)
    return np.dot(v, ww)


class GaussianPsf(object):
    """
    A class for a 2D gaussian PSF

    :math:
       PSF(x,y) = F \times \frac{|W|^{1/2}}{2\pi} \exp(-\frac{1}{2} (X-X_0)^T W (X-X_0))
    """    
    def __init__(self):
        self.pars = self.init_pars()        

    def init_pars(self):
        """Default parameters for a gaussian PSF.

        By default, the PSF is centered and normalized to 1.

        Args:
            seeing (float): seeing value
            fwhm (float, optional): FWHM
        """
        p = FitParameters(['flux', 'x0', 'y0', 'wxx', 'wyy', 'wxy'])
        p['flux'], p['x0'], p['y0'] = 1., 0., 0.
        p['wxx'], p['wyy'], p['wxy'] = 0.25, 0.25, 0.
        return p
    
    @property
    def fwhm(self):
        wxx, wyy = self.pars['wxx'].full[0], self.pars['wyy'].full[0]
        ww = np.sqrt(wxx*wyy)
        return 2. * np.sqrt(2. * np.log(2.) / ww)

    @fwhm.setter
    def fwhm(self, v):
        old_fwhm = self.fwhm
        scale = (old_fwhm / v)**2
        self.pars['wxx'].full *= scale
        self.pars['wyy'].full *= scale
        self.pars['wxy'].full *= scale

    @property
    def sigma(self):
        wxx, wyy = self.pars['wxx'].full[0], self.pars['wyy'].full[0]
        return 1. / np.power(wxx*wyy, 0.25)

    @sigma.setter
    def sigma(self, v):
        old_sigma = self.sigma
        scale = (old_sigma / v) ** 2
        self.pars['wxx'].full *= scale
        self.pars['wyy'].full *= scale
        self.pars['wxy'].full *= scale
        
    def profile(self, x, y, fp=None):
        """PSF profile at positions (x,y)

        Evaluate and return the PSF profile at the specified
        positions. The function is written so that the shape of 
        the input arrays (x,y) is preserved.
        
        Args:
            x (ndarray of floats): x-positions (pixels)
            y (ndarray of floats): y-positions (pixels)
            fp (FitParameters, optional): parameters of the gaussian

        Returns:
            ndarray of floats: the values of the (non integrated) profile
            
        .. note:
            on a Intel Core i5-2540M CPU @ 2.60GHz, about 100 micro-sec 
            to evaluate on ~ 1800 points.
        """
        if fp is None:
            fp = self.pars
        flux, x0, y0, wxx, wyy, wxy = fp.full
        det = wxx*wyy-wxy**2
        norm = np.sqrt(det) / (2.*np.pi)
        xx, yy = x-x0, y-y0
        r = -0.5 * (xx*xx*wxx + yy*yy*wyy + 2.*wxy*xx*yy)
        return flux * norm * np.exp(r)
    
    def __call__(self, x, y, fp=None):
        """psf value, integrated in the pixels.
        
        evaluate the psf value, integrated in each of the pixels 
        where x,y belong.
        
        Args:
            x (ndarray of floats): x-positions (pixels)
            y (ndarray of floats): y-positions (pixels)
            fp (FitParameters, optional): PSF parameters
               if not present, the PSF uses its own internal 
               parameter vector. If is safer to use a PSF
               parameter generated by self.init_pars()

        Returns:
            array of floats: psf values.

        .. note: on a Intel Core i5-2540M CPU @ 2.60GHz, about 750
            micro-sec to evaluate on ~ 1800 points. Almost half of it
            is the call to leggauss in integ_in_pixels
        """
        if fp is not None:
            self.pars.full[:] = fp.full[:]
        if type(x) is float or type(x) is int:
            x = np.array([x])
        if type(y) is float or type(y) is int:
            y = np.array([y])
        return integ_in_pixels(x, y, self.profile)
    
    def pixels(self, rad=5., integrate=True, fp=None, dx=None):
        """array of pixels, containing the PSF values.
        
        Return an array of pixels, with the PSF values. By default, 
        these values are integrated in the pixels. 
        
        Args:
            r (float): radius, in sigma units
            integrate (bool, optional): if true, integrate the PSF profile.
            fp (FitParameters, optional): PSF fit parameters

        Returns:
            ndarray of floats: contains the PSF values
        """
        if fp is None:
            fp = self.pars
        if dx is None:
            dx = int(rad * self.fwhm / 2.)
        x = np.arange(-dx, dx+1., 1.)
        x,y = np.meshgrid(x,x, sparse=True)
        if integrate:
            pix = self(x,y, fp=fp)
        else:
            pix = self.profile(x,y, fp=fp)
        return pix / pix.sum()
        

class MoffatPsf(object):
    """
    A class for a MoffatPSF
    
    :math:
       PSF(X) = \frac{F}{(1 + (X-X_0)^T W (X-X_0))^\alpha}
    """
    def __init__(self, seeing=2., fwhm=None, expn=2.5):
        self.pars = self.init_pars(expn)
        self.fwhm = 4.7096
        
    def init_pars(self, expn):
        """return a initialized parameter vector for the psf. 
        
        Args:
            seeing (float): sigma, in pixels 
            fwhm (float, optional): fwhm, in pixels
            expn (float): Moffat exponent

        Returns:
            FitsParameters: a vector of type FitParameters, 
                initialized so that the PSF has the requested seeing or fwhm
        """
        p = FitParameters(['flux', 'x0', 'y0', 'wxx', 'wyy', 'wxy', 'expn'])
        p['flux'], p['expn'] = 1., expn
        p['wxx'], p['wyy'], p['wxy'] = 0.05761906, 0.05761906, 0.
        return p

    @property
    def fwhm(self):
        expn = self.pars['expn'].full[0]
        wxx, wyy = self.pars['wxx'].full[0], self.pars['wyy'].full[0]
        ww = np.sqrt(wxx*wyy)
        return 2. * np.sqrt((np.power(2., 1/expn) - 1.) / ww)
    
    @fwhm.setter 
    def fwhm(self, v):
        old_fwhm = self.fwhm
        scale = (old_fwhm / v) ** 2
        self.pars['wxx'].full *= scale
        self.pars['wyy'].full *= scale
        self.pars['wxy'].full *= scale        
        
    def profile(self, x, y, fp=None):
        """
        return the (non-integrated) PSF profile for the specified x,y positions
        
        Args:
            x (ndarray of floats): x-positions in pixels
            y (ndarray of floats): y-positions in pixels            
            fp (FitParameters, optional): parameter vector

        Returns:
            the psf profile values at positions x,y

        .. note:
            the shape of x,y is preserved
        """
        if fp is None:
            fp = self.pars
        flux, x0, y0, wxx, wyy, wxy, expn = fp.full
        det = wxx*wyy-wxy**2        
        norm = np.sqrt(np.fabs(det)) * (expn-1.) / np.pi
        xx,yy = x-x0, y-y0
        fact = 1. / (1. + xx*xx*wxx + yy*yy*wyy + 2*xx*yy*wxy)
        return norm * flux * np.power(fact, expn)

    def __call__(self, x, y, fp=None):
        if fp is not None:
            self.pars.full[:] = fp.full[:]
        if type(x) is float or type(x) is int:
            x = np.array([x])
        if type(y) is float or type(y) is int:
            y = np.array([y])
        return integ_in_pixels(x, y, self.profile)

    def pixels(self, rad=5., integrate=True, fp=None, dx=None):
        if fp is None:
            fp = self.pars
        if dx is None:
            dx = int(rad * self.fwhm / 2.)
        x = np.arange(-dx, dx+1., 1.)
        x,y = np.meshgrid(x,x, sparse=True)
        if integrate:
            pix = self(x,y, fp=fp)
        else:
            pix = self.profile(x,y, fp=fp)
        return pix / pix.sum()

def new_psf(name):
    import re
    if name == 'GAUSSIAN':
        return GaussianPsf()
    expr = re.search('MOFFAT(\d+)', name)
    if expr is not None:
        try:
            beta = float(expr.group(1)) * 0.1
            assert beta > 0.
            return MoffatPsf(expn=beta)
        except:
            pass
    return None


class PsfFluxFisher(object):
    """Efficient computation of the variance of a PSF flux estimator.
    
    The variance of the flux is given by: 
    :math:
        V(f) = \left(\sum_i\frac{\psi_i^2}{f \psi_i + f_{sky|i} + \sigma^2_{RON}}\right)^{-1}

    The class precomputes the values of function:
    :math:
        W(x,\sigma) = \sum_i \frac{\psi_i^2}{\psi_i + x}

    over a user-defined range in fwhm, and a range
    :math:`[10^{-6},10^6]` in x. 

    By interpolating (linearly) over the precomputed values one can
    compute very efficiently the variance of the measured PSF flux,
    given the flux (in e-/s), the sky value (e-/s/pixel) and the
    seeing (fwhm, pixels):
    :math:
       V(f) = \frac{f}{W(\frac{f_{sky} + \sigma^2_{RON}}{f}, fwhm)}

    .. note: the class does not know anything about pixel size,
             arcseconds, gain, exposure times and the like. 
             
             Fluxes and sky fluxes are in electrons, seeing is fwhm in
             pixels.  The conversion from physical quantities is
             performed elsewhere (in the instrument class).
    """
    def __init__(self, psf_model=None):
        if psf_model is not None:
            self.psf_model = psf_model
        else:
            self.psf_model = GaussianPsf()

    def precompute(self, x_range=(-6., 6.), fwhm_range=(0.5, 15.), full=False):
        self._precompute_W(x_range=x_range, fwhm_range=fwhm_range)
        if full:
            self._precompute_U()
                    
    def _precompute_W(self, x_range=(-6, 6.), fwhm_range=(0.5, 15.)):
        """precompute the auxiliary function W(x,fwhm)
        
        Loop over the x- and fwhm-ranges, and precompute the W(x,fwhm)
        values over that grid.

        In fact, what is pre-computed and stored is: 
        :math:
           \log_{10} W(\log_10(x), fwhm)

        .. note: This function is somewhat slow (typically, half a second), 
                 but needs to be called only once in the life of the object
                 (unless one decides to change the PSF model, in which case 
                 it has to be called again)
        """
        logging.info('ETC: precomputing W() function [takes a few secs]')
        self.fwhm_range = fwhm_range
        self.x_range = x_range
        fwhm = np.arange(fwhm_range[0], fwhm_range[1], 0.1)
        x = np.logspace(x_range[0], x_range[1], 100)
        x = x.reshape(x.shape + (1,))
        ww, invnea = [], []
        # we loop over the fwhm values first, 
        # to minimize the recomputations of the pixel grids.
        for s in fwhm:
            self.psf_model.fwhm = s
            pix = self.psf_model.pixels(integrate=True).reshape(1, -1)
            val = np.sum(pix**2 / (pix + x), axis=1)
            ww.append(np.log10(val).ravel())
            invnea.append(np.sum(pix**2))
        self.x = x, self.fwhm = x, fwhm
        self.ww = np.vstack(ww)
        self._W = RectBivariateSpline(np.log10(x), fwhm, self.ww.T)
        # need to interpolate between the NEA
        self.invnea = np.asarray(invnea)
        self._invnea = interp1d(fwhm, self.invnea)        

    def _precompute_U(self):
        """
        to be called after precompute_W
        """
        if not hasattr(self, '_W'):
            self._precompute_W()
        logging.info('ETC: precomputing U() function [takes a few secs]')
        from scipy.optimize import brenth
        def f(lx, seeing, lp):
            return self._W.ev(lx,seeing) - lx - lp
        fwhm = np.arange(self.fwhm_range[0], self.fwhm_range[1], 0.1)
        lp = np.linspace(-5., 5., 200)
        gp,gs = np.meshgrid(lp,fwhm)
        data = []
        lx_min, lx_max = -6., 6.
        for lp_,s_ in zip(gp.ravel(),gs.ravel()):
            lx = brenth(f, lx_min, lx_max, args=(s_,lp_))
            data.append(lx)
        self.u_data = data = np.asarray(data)
        self._U = RectBivariateSpline(lp, fwhm, data.reshape(gs.shape).T)
        
    def W(self, x, fwhm):
        """evaluate W for the values of x and fwhm passed as args.

        Evaluate 

        Args:
            x (ndarray of floats): x-values
            fwhm (ndarray of floats): y-values

        Returns:
            ndarray of floats: contains the values of W 
        
        Raises: 
            AssertionError if precompute has not been called.
        """
        assert len(x) == len(fwhm)
        if not hasattr(self, '_W'):
            self._precompute_W()
        x_min, x_max = 10**self.x_range[0], 10**self.x_range[1]
        s_min, s_max = self.fwhm_range
        lx = np.log10(x)
        r = 10**self._W.ev(lx,fwhm)
        # if x outside interp bounds, we have asymptotic values
        r[x<x_min] = 1.
        idx = x>x_max
        r[idx] = self._invnea(fwhm[idx]) / (x[idx])
        # if fwhm outside interp bounds, return very low weight 
        # why not zero ? 
        r[(fwhm<s_min) | (fwhm>s_max)] = 0. # 1.E-9
        return r        
    
    def flux_variance(self, flux, sky, fwhm, ron=5.):
        """compute the PSF flux variance

        Compute the variance of the PSF flux estimator, given values 
        for the flux, sky, seeing (in pixels) and readout noise.
        
        Args:
            flux (ndarray of floats): object fluxes (in electrons)
            sky (ndarray of floats): sky background value (in electrons)
            fwhm (ndarray of floats): fwhm of the PSF (pixels)
            ron (float): readout noise (electrons)
            
        Returns:
            ndarray of floats: flux variance
        """
        fwhm = np.asarray(fwhm)
        flux = np.asarray(flux)
        sky = np.asarray(sky)
        x = (sky + ron**2) / flux
        w = self.W(x, fwhm)
        return flux / w
    
    def noise_equivalent_area(self, fwhm):
        if not hasattr(self, '_invnea'):
            self._precompute_W()
        return 1. / self._invnea(fwhm)

    def signal_to_noise(self, flux, sky, fwhm, ron=5.):
        """return the signal-to-noise ratio

        """
        fwhm = np.asarray(fwhm)
        flux = np.asarray(flux)
        sky = np.asarray(sky)
        x = (sky + ron**2) / flux 
        return np.sqrt(flux * self.W(x, fwhm))

    def flux_lim(self, sky, fwhm, ron=5., snr=5.):
        """return the limiting flux that correspond to the snr specified in argument

        Args:
            sky (ndarray of floats): sky level, in electrons
            fwhm (ndarray of floats): PSF fwhm, in pixels

        Returns:
            ndarray of floats: limiting flux
        """
        if not hasattr(self, '_U'):
            self._precompute_U()
        log_snr_lim = np.log10(snr**2 / (sky+ron**2))
        log_x_lim = self._U.ev(log_snr_lim, fwhm)
        return (sky + ron**2) * 10**(-log_x_lim)
        
    def plot_variance(self, flux_range=(0.1, 1E9), sky=5000., fwhm=4., ron=5., npts=200, 
                      slow=False, fmt='b-'):
        import pylab as pl 
        f_min, f_max = list(map(np.log10, flux_range))
        flux_ = np.logspace(f_min, f_max, npts)
        sky_ = np.ones(npts) * sky
        fwhm_ = np.ones(npts) * fwhm
        if slow is False:
            V = self.flux_variance(flux_, sky_, fwhm_, ron)
        else:
            V = self.flux_variance_slow(flux_, sky_, fwhm_, ron)
        pl.loglog(flux_, V, fmt)
        pl.xlabel('Flux [e-]')
        pl.ylabel('V(Flux)')
        pl.title('PSF flux variance [sky=%6.1f e-, fwhm=%3.1f ron=%4.2f]' % (sky, fwhm, ron))

    def plot_snr(self, flux_range=(0.1, 1.E9), sky=5000., fwhm=4., ron=5., npts=200,
                 slow=True, fmt='b-'):
        import pylab as pl
        f_min, f_max = list(map(np.log10, flux_range))
        flux_ = np.logspace(f_min, f_max, npts)
        sky_ = np.ones(npts) * sky
        fwhm_ = np.ones(npts) * fwhm
        if slow is False:
            snr = self.signal_to_noise(flux_, sky_, fwhm_, ron)
        else:
            sig = np.sqrt(self.flux_variance_slow(flux_, sky_, fwhm_, ron))
            snr = flux_ / sig            
        pl.loglog(flux_, snr, fmt)
        pl.xlabel('Flux [e-]')
        pl.ylabel('SNR(Flux)')        
        pl.title('PSF flux SNR [sky=%6.1f e-, fwhm=%3.1f ron=%4.2f]' % (sky, fwhm, ron))
    
    def plot_fluxlim(self, snr=5., sky_range=(1., 1.E5), fwhm=4., ron=5., npts=200, fmt='b-'):
        import pylab as pl
        sky_min, sky_max = list(map(np.log10, sky_range))
        sky_ = np.logspace(sky_min, sky_max, npts)
        fwhm_ = np.ones(npts) * fwhm
        flux_lim = self.flux_lim(sky_, fwhm_, ron=ron, snr=snr)
        pl.plot(sky_, flux_lim, fmt)
        pl.xlabel('sky level [e-/pixel]')
        pl.ylabel('Flux lim [SNR=%5.1f]' % snr)        
        pl.title('PSF Flux lim [fwhm=%3.1f ron=%4.2f]' % (fwhm, ron))
        
    def flux_variance_slow(self, flux, sky, seeing, ron=5.):
        """compute the PSF flux variance (slow version)

        Same as above, but does not interpolate through precomputed
        values. Use it for debugging purposes only.

        .. note: on a Intel Core i5-2540M CPU @ 2.60GHz, about 
                 2500 times slower than function above. 
        """
        r = []
        for s, f, sk in zip(seeing, flux, sky):
            self.psf_model.fwhm = s
            pix = self.psf_model.pixels(integrate=True).ravel()
            v = np.sum(pix**2 / (f * pix + sk + ron**2))
            r.append(1./v)
        return np.array(r)

    def W_slow(self, fwhm, x):
        r = []
        for s,xx in zip(fwhm, x):
            self.psf_model.fwhm = s
            pix = self.psf_model.pixels(integrate=True).ravel()
            v = np.sum(pix**2 / (pix + xx))
            r.append(1. / v)
        return np.array(r)


class ImagingInstrument(object):
    """Exposure Time Calculator (ETC) for a specific imager.
    
    An instance of ImagingInstrument stores the characteristics of an
    instrument (passbands, pixel size, readout noise) and site
    (seeing, average sky brightness etc.)  Compute flux variance
    and limiting magnitudes for given observing conditions.
    
    An imaging instrument is probably the most 
    
    """
    def __init__(self, instrument_name, **kwargs):
        self.instrument = InstrumentModel(instrument_name)
        cards = self.instrument.cards
        # general cards
        self.ron = self._kw_or_card('ro_noise', kwargs, cards)
        self.dark_current = self._kw_or_card('dark_current', kwargs, cards)
        self.pixel_size = self._kw_or_card('pix_size', kwargs, cards)
        self.pixel_area = self.pixel_size ** 2
        
        # default IQ
        self.iq_model = self._load_default_iq_model()
        
        # default sky background model
        self.sky_background = self._load_default_sky_background()
        
        # psf and etc 
        psf_type = self._kw_or_card('PSF_MODEL', kwargs, cards, default='GAUSSIAN')
        self.psf = new_psf(psf_type)
        self.etc = PsfFluxFisher(self.psf)        
        #        self.etc.precompute(full=1)

        # key data per band 
        self.magsys = MagSys(kwargs.get('magsys', 'AB'))
        self.data = self._compute()
        self.zp = dict(list(zip(self.data['band'], self.data['zp'])))

    def _compute(self):
        fw = self.filter_wheel = self.instrument.filters
        band = list(fw.transmissions.keys())
        tr = [self.instrument.EffectiveFilterByBand(k) for k in band]
        zp = [self.magsys.ZeroPoint(t) for t in tr]
        #        self.zp = dict(zip(band, zp))
        leff = [t.mean(t.x_min, t.x_max) for t in tr]
        iq = [float(self.iq_model(l)) for l in leff]
        msky = [self.magsys.mag(t, self.sky_background)[0][0] for t in tr]
        fsky = [self.sky_background.IntegFlux(t) * self.pixel_area for t in tr]
        #  nea = [self.etc.noise_equivalent_area(s/self.pixel_size) for s in iq]
        #  exptime = [exptime] * len(band)        
        #  m5 = [float(self.mag_lim(t,m,s,b,snr=5.)) for t,m,s,b in zip(exptime,msky,iq,band)]
        full_band_name = [self.instrument.name + '::' + k for k in band]
        ret = np.rec.fromarrays((full_band_name, zp, leff, iq, msky, fsky, tr), 
                                dtype=[('band', '|U20')] + \
                                    [(n,'f8') for n in ('zp', 'leff', 'iq', 'mag_sky', 'flux_sky')] + \
                                    [('tr', 'O')])
        ret.sort(order='leff')
        return ret

    def _kw_or_card(self, kw, kwargs, cards, default=None, raise_exc=False):
        """utility function to interpret the datacards and kw arguments
        """
        r = kwargs.get(kw, None)
        if r is None:
            r = cards.get(kw.upper(), default)
            if r is None and raise_exc:
                raise ValueError('no value specified for: %s' % kw)
        try:
            r = int(r)
        except:
            try:
                return float(r)
            except:
                return r
        return r

    def _load_default_iq_model(self):
        if 'IQ_MODEL' not in self.instrument.cards:
            self.iq = scipy.interpolate.interp1d((3000., 12000.), (1., 1.))
        r = self.instrument.cards['IQ_MODEL'].split()
        if r[0] == 'IQTABLE':
            nm = self.instrument.path + os.sep + r[1]
            d = []
            with open(nm) as f:
                for line in f.readlines():
                    if line[0] == '#': 
                        continue
                    d.append(list(map(float, line.split())))
            iq = np.rec.fromrecords(d, names=['wl', 'iq'])
            return scipy.interpolate.interp1d(iq['wl'], iq['iq'])
        else:
            raise NotImplementedError('model: %s' % r[0])        

    def _load_default_sky_background(self):
        cards = self.instrument.cards
        assert 'SKY_BACKGROUND' in cards
        if cards['SKY_BACKGROUND'] == 'MAUNAKEA_GREY':
            return load_ascii(self.instrument.path + os.sep + '../../maunakea_sky' + os.sep + 'skyspectrum.dat')
        elif cards['SKY_BACKGROUND'] == 'ZODIACAL_LIGHT':
            return load_ascii(self.instrument.path + os.sep + '../../zodiacal_light' + os.sep + 'zodiacal_light.dat')
        else:
            raise ValueError('unable to load sky background')
        
    def _broadcast(self, *args):
        a = tuple([np.asarray(x).reshape(-1) for x in args])
        return np.broadcast_arrays(*a)        

    def precompute(self, full=False):
        self.etc.precompute(full=full)

    def zero_points(self, band):
        return np.asarray([self.zp[b] for b in band])
    
    def flux_to_mag(self, flux, band, zp=None):
        if zp is None:
            zp = self.zero_points(band)
        m = -2.5 * np.log10(flux) + zp
        return m

    def mag_to_flux(self, mag, band, zp=None):
        if zp is None:
            zp = self.zero_points(band)
        return np.power(10., -0.4 * (mag-zp))
        
    def flux_variance(self, exptime, flux, skyflux, seeing, band):
        """The variance of PSF fluxes in the requested bands.

        Args:
          exptime (float or ndarray of floats): exposure times
          flux (float or ndarray of float, optional): source flux on the focal plane (in e/s)
          skyflux (float or ndarray of float, optional): sly brightness (in e/s/arcsecond^2)
          seeing (float or ndarray of floats): PSF FWHM in arcseconds
          band (str or ndarray of str): band

        Returns:
           float or ndarray of floats: flux variance

        .. note :: magnitude system specified in class constructor
        """
        exptime, flux, skyflux, seeing, band = self._broadcast(exptime, flux, skyflux, seeing, band)
        flx = flux * exptime
        flx[flx<0.] = 1.E-6
        v = self.etc.flux_variance(flx, # flux*exptime,
                                   skyflux*exptime*self.pixel_area,
                                   seeing/self.pixel_size,
                                   ron=self.ron)
        return v / exptime**2

    def signal_to_noise(self, exptime, flux, skyflux, seeing, band):
        """The SNR ratios 
        """
        exptime, flux, skyflux, seeing, band = self._broadcast(exptime, flux, skyflux, seeing, band)
        #        zp = np.asarray([self.zp[b] for b in band])
        #        flux = np.power(10., -0.4 * (mag-zp)) * exptime
        #        sky = np.power(10., -0.4 * (skymag-zp)) * self.pixel_area * exptime
        var = self.etc.flux_variance(flux*exptime,
                                     skyflux*exptime*self.pixel_area,
                                     seeing/self.pixel_size,
                                     ron=self.ron)
        return flux / np.sqrt(var)
        
    def flux_lim(self, exptime, skyflux, seeing, band, snr=5.):
        """Limiting fluxes
        """
        exptime, skyflux, seeing, band = self._broadcast(exptime, skyflux, seeing, band)
        #        zp = np.asarray([self.zp[b] for b in band])
        #        sky = np.power(10., -0.4 * (skymag-zp)) * self.pixel_area * exptime
        flim = self.etc.flux_lim(skyflux*exptime*self.pixel_area,
                                 seeing/self.pixel_size,
                                 ron=self.ron,
                                 snr=snr)
        return flim/exptime

    def mag_lim(self, exptime, skyflux, seeing, band, snr=5.):
        """Magnitude limit
        
        """
        exptime, skyflux, seeing, band = self._broadcast(exptime, skyflux, seeing, band)
        zp = np.asarray([self.zp[b] for b in band])
        #        sky = np.power(10., -0.4 * (skymag-zp)) * self.pixel_area * exptime
        flux_lim = self.etc.flux_lim(skyflux*exptime*self.pixel_area, 
                                     seeing/self.pixel_size, 
                                     ron=self.ron, 
                                     snr=snr)
        return -2.5 * np.log10(flux_lim/exptime) + zp
    
    def dump(self, exptime=3600., return_data=False):
        """
        """
        d = self.data
        nea = [self.etc.noise_equivalent_area(s/self.pixel_size) for s in self.data['iq']]
        exptime = [exptime] * len(d)
        skyflux = self.mag_to_flux(d['mag_sky'], d['band'])
        m5 = [float(self.mag_lim(t,m,s,b,snr=5.)) \
                  for t,m,s,b in zip(exptime, skyflux, d['iq'], d['band'])]
        
        print(('-' * 72))
        s = '%-10s  %8s %6s   %6s  %6s  %6s  %6s  %6s %6s' % ('band', 'leff',  'ZP',   'msky',    'fsky',  'FWHM',   'NEA',    'gFWHM', 'm5')
        print(s)
        s = '%-10s  %8s %6s   %6s  %6s  %-6s  %6s  %6s %6s' % ('',     '(AA)',  '(AB)', '(m/"^2)', '(e/px/s)', '(")', '(px^2)', '(")', '')
        print(s)
        print(('-' * 72))
        for i,r in enumerate(self.data):
            s = '%-10s  %8.2f  %6.2f   %6.2f  %6.2f %6.2f  %6.2f  %6.2f %6.2f'  % \
                (r['band'], r['leff'], r['zp'], r['mag_sky'], r['flux_sky'], r['iq'], nea[i], 0., m5[i])
            print(s)
        print(('-' * 72))
        
        if return_data:
            return self.data, np.array(nea), np.array(m5)
        
    def plot(self, wl_range=None):
        import pylab as pl 
        r = self.data['leff'].min()-2000., self.data['leff'].max()+2000.
        wl = np.arange(r[0], r[1], 10.)
        color = [pl.cm.jet(int((x-r[0])/(r[1]-r[0]) * 256.)) for x in self.data['leff']]        
        ma = self.instrument.area
        t = self.instrument.mirror_reflectivity
        pl.plot(wl, t(wl), ls='-', lw=2, color='gray', label='$M(\lambda)$')
        t = t * self.instrument.optics_transmission
        pl.plot(wl, t(wl), ls='--', color='gray', label='$M(\lambda) O(\lambda)$')
        t = t * self.instrument.atm_trans
        pl.plot(wl, t(wl), ls='-.', color='gray', label='$M(\lambda) O(\lambda) A(\lambda)$')
        t = t * self.instrument.ccd_qe
        pl.plot(wl, t(wl), ls='-', color='gray', label='$M(\lambda) O(\lambda) A(\lambda) \epsilon(\lambda)$')
        for i,r in enumerate(self.data):
            pl.plot(wl, self.data['tr'][i](wl)/ma, ls='-', color=color[i])
        if wl_range is not None:
            pl.xlim(wl_range)
        pl.ylim((0., 1.2))
        pl.xlabel('$\lambda\ \mathrm{[nm]}$')
        pl.ylabel('$T(\lambda)$')
        pl.legend(loc=9, ncol=2)
    
    def save(self, filename):
        pass

    def load(self, filename):
        pass


_instrument_repository = {}

def find(instrument_name):
    """
    Search for the requested instrument in a global dictionary.  If
    not found, load the instrument, return it (and update the cache).
    """
    global _instrument_repository
    ret = _instrument_repository.get(instrument_name, None)
    if ret is None:
        ret = ImagingInstrument(instrument_name)
        _instrument_repository[instrument_name] = ret
    return ret

def band_colors(band_names):
    import pylab as pl
    blist = np.unique(band_names)
    instrument_names = np.asarray([r[0] for r in np.core.defchararray.split(blist, b':')])
    instruments = [find(nm) for nm in instrument_names]
    data = np.hstack([instr.data for instr in instruments])
    color = [pl.cm.jet(int((wl-3000) * 256. /(11000-3000.))) for wl in data['leff']]
    return dict(list(zip(data['band'], color)))

def filters(fn):
    pass
    

def polyfit2d(x, y, f, deg):
    """Fit a 2D polynomial on experimental values. 
    
    Given points (x,y) and corresponding measurements (f) fit the 2D
    polynomial of degree deg=(degx, degy). 
    
    Args:
        x,y (ndarrays of floats): positions on the plane 
        f (ndarray of float): measurements
        deg (tuple): degree of the polynomial to fit
            deg = (x-degree, y-degree)
       
    Returns:
        the least-square solution, reshaped so that it is 
        directly usable by polyval2d        
    """
    x = np.asarray(x)
    y = np.asarray(y)
    f = np.asarray(f)
    deg = np.asarray(deg)
    J = polyvander2d(x, y, deg)
    J = J.reshape(-1, J.shape[-1])
    f = f.reshape((J.shape[0],))
    c = np.linalg.lstsq(J,f)[0]
    return c.reshape(deg+1)


        # zp = None
        # if mag is not None and flux is None:
        #     zp = np.asarray([self.zp[b] for b in band])
        #     flux = np.power(10., -0.4 * (mag-zp)) * exptime
        # elif mag is None and flux is not None:
        #     flux *= exptime
        # else:
        #     raise ValueError('specify mag or flux')
        # if skymag is not None and skyflux is None:
        #     if not zp:
        #         zp = np.asarray([self.zp[b] for b in band])
        #     sky = np.power(10., -0.4 * (skymag-zp)) * self.pixel_area * exptime
        # elif skymag is None and skyflux is not None:
        #     sky *= self.pixel_area * exptime
        # else:
        #     raise ValueError('specify either sky mag or sky flux')
        # return self.etc.flux_variance(flux, sky, seeing/self.pixel_size, ron=self.ron)
